package com.android.td5;

import android.os.Handler;
import android.os.Looper;

public class SieveModel {
	// Class for threads in the pool, using a Looper to dispatch Runnable objects
	private class MarkerThread extends Thread {
		// Handler object to queue messages to a given thread in the pool
		public Handler callbackHandler;

		@Override
		public void run() {
			// The callback handler to push marking requests to
			Looper.prepare();
			callbackHandler = new Handler();
			// Wait for marking requests indefinitely
			Looper.loop();
		}
	}

	// Class for the sequential algorithm
	private class MarkSequential implements Runnable {
		@Override
		public void run() {
			int seed = 1;
			
			// Loop until the sieve is over
			while (seed <= currentNumber / 2) {
				// Stop immediately when the user requests another sieve
				if (numberChanged())
					break;

				// Find next prime number
				seed = nextSeed(seed);

				// Mark multiples of seed
				markMultiples(seed);
			}
			
			// Completed, compute and print the number of prime numbers
			countAndPrint();
		}
	}

	// Maximum number of threads in the pool
	public static final int MAX_THREADS = 4;

	// Maximum sieve upper bound
	public static final int MAX_NUMBER = 1000000;

	SieveModel(Sieve sieve) {
		this.sieve = sieve;
	}

	private Sieve sieve;
	
	// Sieve upper bound
	private int number;

	// Protect the sieve upper bound from follow-up user requests
	private int currentNumber;

	// Shared variable holding the number of prime integers found
	private int primes;

	// Marking delay, initially set to 0 (Fastest)
	private int delay = 0;

	// Number of marker threads
	private int numThreads = 1;

	// Array to mark prime numbers
	private boolean isPrime[];

	// Marker thread to compute the sieve
	private MarkerThread sieveThread;

	private void printStatus(String s) {
		sieve.printStatus(s);
	}

	private void printResult(String s) {
		sieve.printResult(s);
	}

	private void countAndPrint() {
		primes = 0;

		for (int i = 2; i <= currentNumber; i++) {
			// Stop immediately when the user requests another sieve
			if (numberChanged())
				return;
			if (isPrime[i])
				primes++;
		}

		printResult("There are " + primes + " prime number"
				+ (primes != 1 ? "s" : "") + " until " + number);
		printStatus("Sieve completed");
	}

	// Invalidate ongoing operations
	public void abortSieve() {
		number = 0;
	}

	public int getNumber() {
		return number;
	}
	
	public void setNumber(int number) {
		this.number = number;
	}
	
	public boolean[] getIsPrime() {
		return isPrime;
	}
	
	public void setIsPrime(boolean[] isPrime) {
		this.isPrime = isPrime;
	}

	// Get the sleep delay for marker thread animations
	public int getDelay() {
		return delay;
	}

	// Update the sleep delay for marker thread animations
	public void setDelay(int delay) {
		this.delay = delay;
	}
	
	// Get the number of Marker threads
	public int getNumThreads() {
		return numThreads;
	}
	
	// Set the number of Marker threads
	public void setNumThreads(int t) {
		abortSieve();
		
		if (t < 1)
			t = 1;
		if (t > MAX_THREADS)
			t = MAX_THREADS;
		
		numThreads = t;
	}
	
	// Return true if the user triggered another sieve
	public boolean numberChanged() {
		return number != currentNumber;
	}

	// Auxiliary function to mark multiples of seed
	private void markMultiples(int seed) {
		printStatus("Marking multiples of " + seed);

		// Step by 2*seed to avoid marking even multiples (already marked as even numbers)
		int step = seed==2 ? 2 : 2*seed;
		// Start from seed*seed to avoid marking multiples of lower integers (already marked)
		int seed2 = seed * seed;
		if (seed2 < 0 || seed2 > currentNumber)
			// Note: seed2 is negative in case of arithmetic overflow
			return;
		
		// Mark odd multiples
		for (int i = seed2; i <= currentNumber; i += step)
			if (isPrime[i]) { // No need to work if already marked
				// Mark i as no n-prime
				isPrime[i] = false;

				// Stop immediately when the user requests another sieve
				if (numberChanged())
					return;
			}
	}

	// Return the next number still marked as (possibly) prime
	private int nextSeed(int seed) {
		if (seed <= 2) // Special case for seed == 1 or seed == 2
			return seed + 1;

		// Only iterate on odd numbers
		for (int i = seed + 2; i <= currentNumber / 2; i += 2)
			if (isPrime[i])
				return i;

		// No need to look any further
		return currentNumber / 2 + 1;
	}

	// Create and start the sieve thread
	public void init() {
		sieveThread = new MarkerThread ();
		sieveThread.start();
	}

	// Compute the sieve of Eratosthenes
	public void sieve() {
		primes = 0; 
		printResult("");
		
		number = sieve.parseNumber();

		// Copy the input value to shadow variable
		// Protect the ongoing sieve from races with a new request
		// And also used to detect new requests (with a different number)
		currentNumber = number;

		// Allocate a new array
		isPrime = new boolean[currentNumber + 1];
		// Assume initially that every integer above 2 is a prime number
		for (int i = 2; i <= currentNumber; i++)
			isPrime[i] = true;
		
		sieveThread.callbackHandler.post(new MarkSequential());
	}
}
